-- factorial in the type language

data Zero
data Succ n

type One   = Succ Zero
type Two   = Succ One
type Three = Succ Two
type Four  = Succ Three

-- dynamic representatives for static Peanos

zero  = undefined :: Zero
one   = undefined :: One
two   = undefined :: Two
three = undefined :: Three
four  = undefined :: Four


-- addition, a la Prolog

class Add a b c | a b -> c where
    add :: a -> b -> c

instance              Add  Zero    b  b
instance Add a b c => Add (Succ a) b (Succ c)


-- multiplication, a la Prolog

class Mul a b c | a b -> c where
    mul :: a -> b -> c

instance                           Mul  Zero    b Zero
instance (Mul a b c, Add b c d) => Mul (Succ a) b d


-- factorial, a la Prolog

class Fac a b | a -> b where
    fac :: a -> b

instance                                Fac  Zero    One
instance (Fac n k, Mul (Succ n) k m) => Fac (Succ n) m


-- value transformation

class Val a where
    value  :: Num b => a -> b
    value' :: Num b => a -> b -> b

    value  x    = value' x 0
    value' x v0 = value x + v0

instance Val Zero where
    value' _ v = v

instance Val a => Val (Succ a) where
    value' (_ :: Succ a) v = value' (undefined :: a) (v+1)

-- show

instance Show Zero where
    show _ = "0"

instance Val (Succ a) => Show (Succ a) where
    showsPrec p = showsPrec p . value
